import logging
import os
from typing import Text, Tuple, Generator, Optional, Dict, List, Set, Any

from rasa.nlu.training_data import TrainingData
from rasa.nlu.training_data.formats import RasaReader
from rasa.nlu.training_data.loading import _guess_format
from rasa.nlu.utils import write_to_file
from rasa.utils.io import read_yaml_file, read_file
from rasax.community import config
from rasax.community.services.role_service import RoleService
from rasax.community.services.user_service import UserService

logger = logging.getLogger(__name__)


def inject_config(config_path: Text, settings_service: "SettingsService"):
    """Load a configuration file from `path` and save it to the database."""

    if not os.path.exists(config_path):
        logger.warning(
            "Failed to inject Rasa configuration. The file '{}' "
            "does not exists."
            "".format(os.path.abspath(config_path))
        )
        return

    yaml_config = read_yaml_file(config_path)
    if not yaml_config:
        logger.warning(
            "Failed to inject Rasa configuration:\n"
            "Reading of yaml '{}' file failed. Most likely the "
            "file was not found or uses the wrong syntax."
            "".format(os.path.abspath(config_path))
        )

    stack_config = settings_service.save_config(
        config.team_name, "default", yaml_config, config_path
    )
    if stack_config:
        logger.debug(
            "Loaded local configuration from '{}' into "
            "database".format(os.path.abspath(config_path))
        )
    else:
        logger.warning(
            "Failed to inject Rasa configuration. The file '{}'"
            "was found but could not be saved to the database."
            "".format(os.path.abspath(config_path))
        )


def _read_data(paths: List[Text]) -> Generator[Tuple[Text, Text], None, None]:
    for filename in paths:
        yield read_file(filename), filename


def inject_nlu_data(
    nlu_files: Set[Text], project_id: Text, username: Text, data_service: "DataService"
) -> TrainingData:
    """Load Rasa NLU training data from `path` and save it to the database."""

    # delete existing data in db if files are provided
    if nlu_files:
        data_service.delete_data()

    training_data = TrainingData()

    for data, path in _read_data(list(nlu_files)):
        logger.debug("Injecting NLU data from file '{}' to database.".format(path))

        data_format = _guess_format(path)
        if data_format not in {"md", "json"}:
            logger.debug(
                "Invalid data format for file '{}'. "
                "Accepting 'md' and 'json' but found "
                "'{}'.".format(path, data_format)
            )
            continue

        additional_data = data_service.save_bulk_data(
            project_id,
            data,
            data_format,
            username=username,
            filename=path,
            dump_data_in_local_mode=False,
        )

        if additional_data:
            training_data = training_data.merge(additional_data)

    logger.debug(
        "Injected {} NLU training data examples."
        "".format(len(training_data.training_examples))
    )
    return training_data


def dump_nlu_data(
    data_service: "DataService", project: Text = config.project_name
) -> None:
    """Dump Rasa NLU data in database to file at `path`."""

    data_filenames = data_service.get_all_filenames(project)
    for fname in data_filenames:
        path = os.path.join(fname)
        logger.debug("Dumping NLU data to file '{}'.".format(path))

        nlu_format = data_service.create_formatted_training_data(fname, project)

        training_data = RasaReader().read_from_json(nlu_format)
        write_to_file(path, training_data.as_markdown())


async def inject_stories(
    story_files: Set[Text],
    story_service: "StoryService",
    username: Text,
    team: Text = config.team_name,
) -> List[Dict[Text, Any]]:
    """Load Core stories from `data_directory` and save to the database."""

    # delete existing data in db if files are provided
    if story_files:
        await story_service.delete_stories(team)

    story_blocks = []
    additional_blocks = []

    for data, path in _read_data(list(story_files)):
        logger.debug("Injecting stories from file '{}' to database.".format(path))
        additional_blocks = await story_service.save_stories(
            data, team, path, username, dump_stories_in_local_mode=False
        )
        story_blocks.extend(additional_blocks)

    logger.debug("Injected {} Core stories.".format(len(additional_blocks)))
    return story_blocks


def inject_domain(
    domain_path: Text,
    projects_service: "ProjectsService",
    project_id: Text = config.project_name,
    username: Text = config.default_username,
) -> Dict[Text, Any]:
    """Load Rasa Core domain at `path` and save it to database."""

    if not os.path.exists(domain_path):
        logger.warning(
            "Failed to inject domain file. The file '{}' "
            "does not exists."
            "".format(os.path.abspath(domain_path))
        )
        return {}

    data = read_yaml_file(domain_path)

    # inject domain including templates
    projects_service.store_domain(
        data, project_id, path=domain_path, store_templates=True, username=username
    )

    return data


def inject_user(
    user_service: "UserService",
    role_service: "RoleService",
    username: Text,
    password: Text,
    project_id: Text,
    team: Text,
) -> None:
    """Create a local user with admin privileges."""

    role_service.init_roles(project_id=project_id)

    if user_service.fetch_user(username):
        logger.debug("Found user in local mode: '{}'".format(username))
        user_service.admin_change_password(username, password)
        logger.debug(
            "Updated password for local user {} to {}".format(username, password)
        )
    else:
        user_service.create_user(username, password, team, "admin")
        logger.debug(
            "Created local user named {} with password {}".format(username, password)
        )


def inject_project(
    settings_service: "SettingsService",
    user_service: "UserService",
    project_id: Text,
    team: Text,
    is_in_local_mode: bool,
    user: Optional[Dict] = None,
) -> None:
    """Create project and settings."""

    existing = settings_service.get(team, project_id)
    if existing is None:
        if user is not None:
            user_service.create_project(user, project_id)
        settings_service.init_project(
            team, project_id, is_in_local_mode=is_in_local_mode
        )
        logger.debug("Created project '{}'.".format(project_id))
    else:
        logger.debug("Found project '{}'.".format(project_id))


def create_project_and_settings(_settings_service, _role_service, team):
    project_id = config.project_name

    existing = _settings_service.get(team, project_id)

    if existing is None:
        _settings_service.init_project(team, project_id)

    _role_service.init_roles(project_id=project_id)
