from rasax.community.config import team_name
from rasax.community.services.nlg_service import NlgService
from rasax.community.services.projects_service import ProjectsService
from rasax.community.services.settings_service import SettingsService
from tests.unit.conftest import TEST_USER

TEST_DOMAIN = {
    "actions": ["utter_1", "utter_2"],
    "config": {"store_entities_as_slots": True},
    "entities": [],
    "forms": ["test_form"],
    "intents": [
        {"intent_1": {"use_entities": True}},
        {"intent_2": {"use_entities": False}},
    ],
    "slots": {
        "slot_1": {"auto_fill": True, "type": "rasa_core.slots.UnfeaturizedSlot"},
        "slot_2": {
            "auto_fill": True,
            "initial_value": "value1",
            "type": "rasa.core.slots.CategoricalSlot",
            "values": ["value1", "value2"],
        },
    },
    "templates": {"utter_1": [{"text": "hello world"}]},
}

TEST_PROJECT = "project"


def test_storing_of_domain(
    projects_service: ProjectsService, settings_service: SettingsService
):
    settings_service.init_project(team_name, TEST_PROJECT)

    projects_service.store_domain(
        TEST_DOMAIN, TEST_PROJECT, username=TEST_USER, store_templates=True
    )

    actual = projects_service.get_domain(TEST_PROJECT)

    assert actual == TEST_DOMAIN

    # Save another domain for the same project and check whether old one
    # gets replaced correctly
    TEST_DOMAIN["actions"].append("utter_3")
    projects_service.store_domain(
        TEST_DOMAIN, TEST_PROJECT, username=TEST_USER, store_templates=True
    )

    actual = projects_service.get_domain(TEST_PROJECT)

    assert actual == TEST_DOMAIN


def test_upload_domain_without_templates(
    projects_service: ProjectsService, nlg_service: NlgService
):
    # there are templates
    actual, _ = nlg_service.fetch_templates()
    assert actual

    # upload domain without templates
    domain_without_templates = TEST_DOMAIN.copy()
    domain_without_templates["templates"] = {}
    projects_service.store_domain(
        domain_without_templates, TEST_PROJECT, username=TEST_USER, store_templates=True
    )

    assert projects_service.get_domain(TEST_PROJECT) == domain_without_templates

    # there should be no templates
    actual, _ = nlg_service.fetch_templates()
    assert not actual


def test_store_domain_replaces_templates(
    projects_service: ProjectsService, nlg_service: NlgService
):
    domain_without_templates = TEST_DOMAIN.copy()
    domain_without_templates["templates"] = {}
    projects_service.store_domain(
        domain_without_templates, TEST_PROJECT, username=TEST_USER, store_templates=True
    )

    assert projects_service.get_domain(TEST_PROJECT) == domain_without_templates

    # insert some templates and make sure they end up in domain
    from tests.unit.test_nlg_service import TEST_TEMPLATE

    domain_id = projects_service.get_domain_id(TEST_PROJECT)
    nlg_service.save_template(TEST_TEMPLATE, TEST_USER, domain_id, TEST_PROJECT)

    # get domain again and check template is there
    actual = projects_service.get_domain(TEST_PROJECT)
    assert len(actual["templates"]) == 1
    assert (
        actual["templates"][TEST_TEMPLATE["template"]][0]["text"]
        == TEST_TEMPLATE["text"]
    )


def test_get_intents(
    projects_service: ProjectsService, settings_service: SettingsService
):
    test_project = "project2"
    settings_service.init_project(team_name, test_project)
    test_domain = {
        "intents": [
            {"intent_1": {"use_entities": True}},
            {"intent_2": {"use_entities": False}},
        ]
    }

    projects_service.store_domain(test_domain, test_project)
    actual = projects_service.get_intents_from_domain(test_project)

    assert actual == {"intent_1", "intent_2"}


def test_save_and_get_token(projects_service: ProjectsService):
    from uuid import uuid4

    # sql_migration successfully created token using save_token
    assert len(projects_service.get_token().token) == len(uuid4().hex)
    assert len(str(projects_service.get_token().expires)) == 10


def test_update_token(projects_service: ProjectsService):
    bot_name = "Ursula"
    description = "Bot that will insult you in many creative ways, be prepared!"
    projects_service.update_token(bot_name, description)
    token = projects_service.get_token()
    assert token.bot_name == bot_name
    assert token.description == description
